'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var tslib = require('tslib');

var ValidationError = /** @class */ (function (_super) {
    tslib.__extends(ValidationError, _super);
    function ValidationError(issues) {
        var _newTarget = this.constructor;
        var _this = _super.call(this) || this;
        _this.issues = issues;
        Object.setPrototypeOf(_this, _newTarget.prototype);
        _this.name = 'ValidationError';
        return _this;
    }
    return ValidationError;
}(Error));

function raiseIssue(input, code, param, options, message) {
    var _a;
    raiseIssuesIfDefined([
        {
            code: code,
            path: [],
            input: input,
            param: param,
            message: (_a = options === null || options === void 0 ? void 0 : options.message) !== null && _a !== void 0 ? _a : message,
            meta: options === null || options === void 0 ? void 0 : options.meta,
        },
    ]);
}
function raiseIssuesOrPush(issues, parserOptions, input, code, param, options, message) {
    var _a;
    if (issues === void 0) { issues = []; }
    issues.push({
        code: code,
        path: [],
        input: input,
        param: param,
        message: (_a = options === null || options === void 0 ? void 0 : options.message) !== null && _a !== void 0 ? _a : message,
        meta: options === null || options === void 0 ? void 0 : options.meta,
    });
    if (parserOptions === null || parserOptions === void 0 ? void 0 : parserOptions.fast) {
        raiseIssuesIfDefined(issues);
    }
    return issues;
}
function raiseIssuesIfDefined(issues) {
    if (issues !== undefined) {
        throw new ValidationError(issues);
    }
}
function raiseIssuesOrCaptureForKey(error, issues, parserOptions, key) {
    if (error instanceof ValidationError) {
        for (var _i = 0, _a = error.issues; _i < _a.length; _i++) {
            var issue = _a[_i];
            issue.path.unshift(key);
        }
    }
    if (isFast(parserOptions) || !(error instanceof ValidationError)) {
        throw error;
    }
    if (issues !== undefined) {
        issues.push.apply(issues, error.issues);
        return issues;
    }
    return error.issues;
}
function cloneObject(value) {
    return Object.assign(Object.create(Object.getPrototypeOf(value)), value);
}
function isObjectLike(value) {
    return value !== null && typeof value === 'object';
}
function createCatchForKey(key) {
    return function (error) {
        if (error instanceof ValidationError) {
            for (var _i = 0, _a = error.issues; _i < _a.length; _i++) {
                var issue = _a[_i];
                issue.path.unshift(key);
            }
        }
        throw error;
    };
}
function createValuesExtractor(issues) {
    return function (results) {
        for (var i = 0; i < results.length; ++i) {
            var result = results[i];
            if (result.status === 'rejected') {
                var error = result.reason;
                if (!(error instanceof ValidationError)) {
                    throw error;
                }
                if (issues === undefined) {
                    issues = error.issues;
                }
                else {
                    issues.push.apply(issues, error.issues);
                }
            }
        }
        if (issues !== undefined) {
            throw new ValidationError(issues);
        }
        var values = [];
        for (var i = 0; i < results.length; ++i) {
            values.push(results[i].value);
        }
        return values;
    };
}
function isFast(options) {
    return options !== undefined && options.fast === true;
}
function isAsync(types) {
    var async = false;
    for (var i = 0; i < types.length && !async; ++i) {
        async = types[i].async;
    }
    return async;
}
var isEqual = Object.is;
var isArray = Array.isArray;
var promiseAll = Promise.all.bind(Promise);
var promiseAllSettled = Promise.allSettled.bind(Promise);
function parseAsync(type, input, options) {
    return type.async ? type.parse(input, options) : Promise.resolve().then(function () { return type.parse(input, options); });
}
function isEqualArray(a, b) {
    for (var i = 0; i < a.length; ++i) {
        if (!isEqual(a[i], b[i])) {
            return false;
        }
    }
    return true;
}
function returnNull() {
    return null;
}
function extractIssues(error) {
    if (error instanceof ValidationError) {
        return error.issues;
    }
    throw error;
}
function copyObjectKnownKeys(input, keys) {
    var output = {};
    for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
        var key = keys_1[_i];
        output[key] = input[key];
    }
    return output;
}
function copyObjectEnumerableKeys(input, keyCount) {
    var output = {};
    if (keyCount === undefined) {
        for (var key in input) {
            output[key] = input[key];
        }
        return output;
    }
    var i = 0;
    for (var key in input) {
        if (i >= keyCount) {
            break;
        }
        output[key] = input[key];
        i++;
    }
    return output;
}

/**
 * The abstract type definition.
 */
var Type = /** @class */ (function () {
    function Type(async, options) {
        this.async = async;
        this.options = options;
    }
    Type.prototype.validate = function (input, options) {
        if (this.async) {
            throw new Error('Cannot use async type');
        }
        try {
            this.parse(input, options);
        }
        catch (error) {
            return extractIssues(error);
        }
        return null;
    };
    Type.prototype.validateAsync = function (input, options) {
        return parseAsync(this, input, options).then(returnNull, extractIssues);
    };
    /**
     * Transforms the input to the new type.
     *
     * @param transformer The callback that takes the input and transforms it to the new type.
     * @returns The transformed type.
     *
     * @template O The type of the output input.
     */
    Type.prototype.transform = function (transformer) {
        return new TransformedType(this, false, transformer);
    };
    /**
     * Asynchronously transforms the input to the new type.
     *
     * @param transformer The callback that takes the input and transforms it to the new type.
     * @returns The transformed type.
     *
     * @template O The type of the output input.
     */
    Type.prototype.transformAsync = function (transformer) {
        return new TransformedType(this, true, transformer);
    };
    Type.prototype.refine = function (refiner, options) {
        return this.transform(function (input) {
            if (!refiner(input)) {
                raiseIssue(input, 'refined', undefined, options, 'Must be refined');
            }
            return input;
        });
    };
    return Type;
}());
/**
 * The transforming type definition.
 *
 * @template X The type definition of the input value.
 * @template T The output value.
 */
var TransformedType = /** @class */ (function (_super) {
    tslib.__extends(TransformedType, _super);
    /**
     * Creates a new {@link TransformedType} instance.
     *
     * @param type The type that parses a transformation input value.
     * @param async `true` if transformer returns a `Promise`, or `false` otherwise.
     * @param transformer The transformer that converts input value to the output value.
     */
    function TransformedType(type, async, transformer) {
        var _this = _super.call(this, async || type.async) || this;
        _this.type = type;
        _this.transformer = transformer;
        return _this;
    }
    TransformedType.prototype.parse = function (input, options) {
        var _a = this, type = _a.type, transformer = _a.transformer;
        if (this.async) {
            var promise = type.async ? type.parse(input, options) : parseAsync(type, input, options);
            return promise.then(transformer);
        }
        return transformer(type.parse(input, options));
    };
    return TransformedType;
}(Type));

/**
 * The array type definition.
 *
 * @template X The type definition of array elements.
 */
var ArrayType = /** @class */ (function (_super) {
    tslib.__extends(ArrayType, _super);
    /**
     * Creates the new {@link ArrayType} instance.
     *
     * @param type The type definition of array elements.
     * @param options The type constraint options.
     */
    function ArrayType(type, options) {
        var _this = _super.call(this, type.async, options) || this;
        _this.type = type;
        return _this;
    }
    /**
     * Constrains the array length.
     *
     * @param length The length of the array to satisfy the constraint.
     * @param options The options applied to both {@link min} and {@link max} constraints.
     * @returns The new type definition.
     */
    ArrayType.prototype.length = function (length, options) {
        return this.min(length, options).max(length, options);
    };
    /**
     * Constrains the minimum array length.
     *
     * @param length The minimum length of the array to satisfy the constraint.
     * @param options The minimum constraint options.
     * @returns The new type definition.
     */
    ArrayType.prototype.min = function (length, options) {
        var type = cloneObject(this);
        type.minLength = length;
        type.minLengthOptions = options;
        return type;
    };
    /**
     * Constrains the maximum array length.
     *
     * @param length The maximum length of the array to satisfy the constraint.
     * @param options The maximum constraint options.
     * @returns The new type definition.
     */
    ArrayType.prototype.max = function (length, options) {
        var type = cloneObject(this);
        type.maxLength = length;
        type.maxLengthOptions = options;
        return type;
    };
    ArrayType.prototype.parse = function (input, options) {
        if (!isArray(input)) {
            raiseIssue(input, 'type', 'array', this.options, 'Must be an array');
        }
        var _a = this, minLength = _a.minLength, maxLength = _a.maxLength, type = _a.type;
        var inputLength = input.length;
        var issues;
        if (minLength != null && inputLength < minLength) {
            issues = raiseIssuesOrPush(issues, options, input, 'arrayMinLength', minLength, this.minLengthOptions, 'Must have the minimum length of ' + minLength);
        }
        if (maxLength != null && inputLength > maxLength) {
            issues = raiseIssuesOrPush(issues, options, input, 'arrayMaxLength', maxLength, this.maxLengthOptions, 'Must have the maximum length of ' + maxLength);
        }
        if (this.async) {
            var promises = [];
            var handleOutput = function (output) { return (isEqualArray(input, output) ? input : output); };
            for (var i = 0; i < inputLength; ++i) {
                promises.push(type.parse(input[i], options).catch(createCatchForKey(i)));
            }
            if (isFast(options)) {
                return promiseAll(promises).then(handleOutput);
            }
            return promiseAllSettled(promises).then(createValuesExtractor(issues)).then(handleOutput);
        }
        var output = input;
        for (var i = 0; i < inputLength; ++i) {
            var value = input[i];
            var outputValue = void 0;
            try {
                outputValue = type.parse(value, options);
            }
            catch (error) {
                issues = raiseIssuesOrCaptureForKey(error, issues, options, i);
            }
            if (isEqual(outputValue, value) || issues !== undefined) {
                continue;
            }
            if (output === input) {
                output = input.slice(0);
            }
            output[i] = outputValue;
        }
        raiseIssuesIfDefined(issues);
        return output;
    };
    return ArrayType;
}(Type));

/**
 * The bigint type definition.
 */
var BigIntType = /** @class */ (function (_super) {
    tslib.__extends(BigIntType, _super);
    function BigIntType(options) {
        return _super.call(this, false, options) || this;
    }
    BigIntType.prototype.parse = function (input, options) {
        if (typeof input !== 'bigint') {
            raiseIssue(input, 'type', 'bigint', this.options, 'Must be a bigint');
        }
        return input;
    };
    return BigIntType;
}(Type));

/**
 * The boolean type definition.
 */
var BooleanType = /** @class */ (function (_super) {
    tslib.__extends(BooleanType, _super);
    function BooleanType(options) {
        return _super.call(this, false, options) || this;
    }
    BooleanType.prototype.parse = function (input, options) {
        if (typeof input !== 'boolean') {
            raiseIssue(input, 'type', 'boolean', this.options, 'Must be a boolean');
        }
        return input;
    };
    return BooleanType;
}(Type));

/**
 * The class instance type definition.
 *
 * @template C The class instance constructor.
 */
var InstanceOfType = /** @class */ (function (_super) {
    tslib.__extends(InstanceOfType, _super);
    function InstanceOfType(ctor, options) {
        var _this = _super.call(this, false, options) || this;
        _this.ctor = ctor;
        return _this;
    }
    InstanceOfType.prototype.parse = function (input, options) {
        var ctor = this.ctor;
        if (!(input instanceof ctor)) {
            raiseIssue(input, 'instanceOf', ctor, this.options, 'Must be an instance of ' + ctor.name);
        }
        return input;
    };
    return InstanceOfType;
}(Type));

var isFinite = Number.isFinite;
/**
 * The number type definition.
 */
var NumberType = /** @class */ (function (_super) {
    tslib.__extends(NumberType, _super);
    function NumberType(options) {
        return _super.call(this, false, options) || this;
    }
    /**
     * Constrains the number to be greater than zero.
     */
    NumberType.prototype.positive = function (options) {
        return this.gt(0, options);
    };
    /**
     * Constrains the number to be less than zero.
     */
    NumberType.prototype.negative = function (options) {
        return this.lt(0, options);
    };
    /**
     * Constrains the number to be greater than the value.
     */
    NumberType.prototype.gt = function (value, options) {
        var type = cloneObject(this);
        type.min = value;
        type.minIncluded = false;
        type.minOptions = options;
        return type;
    };
    /**
     * Constrains the number to be less than the value.
     */
    NumberType.prototype.lt = function (value, options) {
        var type = cloneObject(this);
        type.max = value;
        type.maxIncluded = false;
        type.maxOptions = options;
        return type;
    };
    /**
     * Constrains the number to be greater than or equal to the value.
     */
    NumberType.prototype.gte = function (value, options) {
        var type = cloneObject(this);
        type.min = value;
        type.minIncluded = true;
        type.minOptions = options;
        return type;
    };
    /**
     * Constrains the number to be less than or equal to the value.
     */
    NumberType.prototype.lte = function (value, options) {
        var type = cloneObject(this);
        type.max = value;
        type.maxIncluded = true;
        type.maxOptions = options;
        return type;
    };
    /**
     * Constrains the number to be a multiple of the divisor.
     *
     * @param divisor The number by which the input should be divisible without a remainder.
     * @param options
     */
    NumberType.prototype.multipleOf = function (divisor, options) {
        var type = cloneObject(this);
        type.divisor = divisor;
        type.divisorOptions = options;
        return type;
    };
    NumberType.prototype.parse = function (input, options) {
        if (!isFinite(input)) {
            raiseIssue(input, 'type', 'number', this.options, 'Must be a number');
        }
        var _a = this, min = _a.min, max = _a.max, minIncluded = _a.minIncluded, maxIncluded = _a.maxIncluded, divisor = _a.divisor;
        var issues;
        if (min != null && (minIncluded ? input < min : input <= min)) {
            issues = raiseIssuesOrPush(issues, options, input, minIncluded ? 'numberGreaterThanOrEqual' : 'numberGreaterThan', min, this.minOptions, 'Must be greater than ' + (minIncluded ? 'or equal to ' + min : min));
        }
        if (max != null && (maxIncluded ? input > max : input >= max)) {
            issues = raiseIssuesOrPush(issues, options, input, maxIncluded ? 'numberLessThanOrEqual' : 'numberLessThan', max, this.maxOptions, 'Must be less than ' + (maxIncluded ? 'or equal to ' + max : max));
        }
        if (divisor != null && input % divisor !== 0) {
            issues = raiseIssuesOrPush(issues, options, input, 'numberMultipleOf', divisor, this.divisorOptions, 'Must be a multiple of ' + divisor);
        }
        raiseIssuesIfDefined(issues);
        return input;
    };
    return NumberType;
}(Type));

var isInteger = Number.isInteger;
/**
 * The integer type definition.
 */
var IntegerType = /** @class */ (function (_super) {
    tslib.__extends(IntegerType, _super);
    function IntegerType(options) {
        return _super.call(this, options) || this;
    }
    IntegerType.prototype.parse = function (input, options) {
        if (!isInteger(input)) {
            raiseIssue(input, 'type', 'integer', this.options, 'Must be an integer');
        }
        return _super.prototype.parse.call(this, input, options);
    };
    return IntegerType;
}(NumberType));

/**
 * The lazily-evaluated type definition.
 *
 * @template X The type definition returned by the provider.
 */
var LazyType = /** @class */ (function (_super) {
    tslib.__extends(LazyType, _super);
    /**
     * Creates a new {@link LazyType} instance.
     *
     * @param provider Returns the type definition that must be applied to the input value.
     */
    function LazyType(async, provider) {
        var _this = _super.call(this, async) || this;
        _this.provider = provider;
        return _this;
    }
    LazyType.prototype.parse = function (input, options) {
        return (this.type || (this.type = this.provider())).parse(input, options);
    };
    return LazyType;
}(Type));

/**
 * The literal value type definition.
 *
 * @template T The literal value.
 */
var LiteralType = /** @class */ (function (_super) {
    tslib.__extends(LiteralType, _super);
    /**
     * Creates a new {@link LiteralType} instance.
     *
     * @param value The literal value that is compared with the input value.
     * @param options
     */
    function LiteralType(value, options) {
        var _this = _super.call(this, false, options) || this;
        _this.value = value;
        return _this;
    }
    LiteralType.prototype.parse = function (input, options) {
        var value = this.value;
        if (!isEqual(input, value)) {
            raiseIssue(input, 'literal', value, this.options, 'Must be exactly equal to ' + value);
        }
        return input;
    };
    return LiteralType;
}(Type));

/**
 * The type definition that always raises an issue.
 */
var NeverType = /** @class */ (function (_super) {
    tslib.__extends(NeverType, _super);
    function NeverType(options) {
        return _super.call(this, false, options) || this;
    }
    NeverType.prototype.parse = function (input, options) {
        raiseIssue(input, 'never', undefined, this.options, 'Must not be used');
    };
    return NeverType;
}(Type));

/**
 * The nullable type definition.
 *
 * @template X The underlying type definition.
 */
var NullableType = /** @class */ (function (_super) {
    tslib.__extends(NullableType, _super);
    /**
     * Creates a new {@link NullableType} instance.
     *
     * @param type The underlying type definition.
     */
    function NullableType(type) {
        var _this = _super.call(this, type.async) || this;
        _this.type = type;
        return _this;
    }
    NullableType.prototype.parse = function (input, options) {
        var type = this.type;
        if (input === null) {
            return type.async ? Promise.resolve(input) : input;
        }
        return type.parse(input, options);
    };
    return NullableType;
}(Type));

/**
 * The object type definition.
 *
 * @template P The mapping from an object key to a corresponding type definition.
 * @template I The type definition that constrains the indexer signature.
 */
var ObjectType = /** @class */ (function (_super) {
    tslib.__extends(ObjectType, _super);
    /**
     * Creates a new {@link ObjectType} instance.
     *
     * @param props The mapping from an object key to a corresponding type definition.
     * @param indexerType The type definition that constrains the indexer signature. If `null` then values thea fall
     * under the indexer signature are unconstrained.
     * @param options
     */
    function ObjectType(props, indexerType, options) {
        var _this = this;
        var valueTypes = Object.values(props);
        _this = _super.call(this, (indexerType === null || indexerType === void 0 ? void 0 : indexerType.async) || isAsync(valueTypes), options) || this;
        _this.props = props;
        _this.indexerType = indexerType;
        _this.keys = Object.keys(props);
        _this.valueTypes = valueTypes;
        _this.propEntries = Object.entries(props);
        _this.keysMode = indexerType === null ? 2 /* ObjectKeysMode.PRESERVE */ : 3 /* ObjectKeysMode.INDEXER */;
        return _this;
    }
    ObjectType.prototype.extend = function (arg) {
        var nextProps = Object.assign({}, this.props, arg instanceof ObjectType ? arg.props : arg);
        var type = new ObjectType(nextProps, this.indexerType);
        type.keysMode = this.keysMode;
        return type;
    };
    /**
     * Returns an object type that only has properties with listed keys.
     *
     * @param keys The list of property keys to pick.
     * @returns The modified object type.
     *
     * @template K The tuple of keys to pick.
     */
    ObjectType.prototype.pick = function () {
        var keys = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            keys[_i] = arguments[_i];
        }
        var nextProps = {};
        for (var _a = 0, keys_1 = keys; _a < keys_1.length; _a++) {
            var key = keys_1[_a];
            nextProps[key] = this.props[key];
        }
        var type = new ObjectType(nextProps, this.indexerType);
        type.keysMode = this.keysMode;
        return type;
    };
    /**
     * Returns an object type that doesn't have the listed keys.
     *
     * @param keys2 The list of property keys to omit.
     * @returns The modified object type.
     *
     * @template K The tuple of keys to omit.
     */
    ObjectType.prototype.omit = function () {
        var keys2 = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            keys2[_i] = arguments[_i];
        }
        var keys = this.keys;
        var nextProps = {};
        for (var i = 0; i < keys.length; ++i) {
            if (!keys2.includes(keys[i])) {
                nextProps[keys[i]] = this.valueTypes[i];
            }
        }
        var type = new ObjectType(nextProps, this.indexerType);
        type.keysMode = this.keysMode;
        return type;
    };
    /**
     * Returns an object type that allows only known keys and has no index signature.
     *
     * @returns The modified object type.
     */
    ObjectType.prototype.exact = function (options) {
        var type = cloneObject(this);
        type.keysMode = 0 /* ObjectKeysMode.EXACT */;
        type.exactOptions = options;
        return type;
    };
    /**
     * Returns an object type that doesn't have indexer signature and all unknown keys are stripped.
     *
     * @returns The modified object type.
     */
    ObjectType.prototype.strip = function () {
        var type = cloneObject(this);
        type.keysMode = 1 /* ObjectKeysMode.STRIP */;
        return type;
    };
    /**
     * Returns an object type that has an indexer signature that doesn't constrain values.
     *
     * @returns The modified object type.
     */
    ObjectType.prototype.preserve = function () {
        var type = cloneObject(this);
        type.keysMode = 2 /* ObjectKeysMode.PRESERVE */;
        return type;
    };
    /**
     * Returns an object type that has an indexer signature that is constrained by the given type.
     *
     * @param indexType The type of the indexer values.
     * @returns The modified object type.
     *
     * @template I1 The indexer signature type.
     */
    ObjectType.prototype.index = function (indexType) {
        var type = cloneObject(this);
        type.keysMode = 3 /* ObjectKeysMode.INDEXER */;
        type.indexerType = indexType;
        return type;
    };
    ObjectType.prototype.parse = function (input, options) {
        if (!isObjectLike(input)) {
            raiseIssue(input, 'type', 'object', this.options, 'Must be an object');
        }
        var _a = this, propEntries = _a.propEntries, keysMode = _a.keysMode, keys = _a.keys, indexerType = _a.indexerType;
        var output = input;
        var issues;
        if (keysMode === 0 /* ObjectKeysMode.EXACT */ || keysMode === 1 /* ObjectKeysMode.STRIP */) {
            for (var key in input) {
                if (keys.includes(key)) {
                    continue;
                }
                if (keysMode === 0 /* ObjectKeysMode.EXACT */) {
                    issues = raiseIssuesOrPush(issues, options, input, 'unknownKeys', key, this.exactOptions, 'Must have known keys but found ' + key);
                }
                if (output === input) {
                    output = copyObjectKnownKeys(input, keys);
                }
            }
        }
        if (this.async) {
            var promises = [];
            var objectKeys_1 = keys;
            var handleResults = function (results) {
                for (var i = 0; i < objectKeys_1.length; ++i) {
                    var key = objectKeys_1[i];
                    var outputValue = results[i];
                    if (isEqual(outputValue, input[key])) {
                        continue;
                    }
                    if (output === input) {
                        output = copyObjectEnumerableKeys(input);
                    }
                    output[key] = outputValue;
                }
                return output;
            };
            for (var _i = 0, propEntries_1 = propEntries; _i < propEntries_1.length; _i++) {
                var _b = propEntries_1[_i], key = _b[0], type = _b[1];
                promises.push(parseAsync(type, input[key], options).catch(createCatchForKey(key)));
            }
            if (keysMode === 3 /* ObjectKeysMode.INDEXER */) {
                for (var key in input) {
                    if (keys.includes(key)) {
                        continue;
                    }
                    if (objectKeys_1 === keys) {
                        objectKeys_1 = keys.slice(0);
                    }
                    objectKeys_1.push(key);
                    promises.push(parseAsync(indexerType, input[key], options).catch(createCatchForKey(key)));
                }
            }
            if (isFast(options)) {
                return promiseAll(promises).then(handleResults);
            }
            return promiseAllSettled(promises).then(createValuesExtractor(issues)).then(handleResults);
        }
        for (var _c = 0, propEntries_2 = propEntries; _c < propEntries_2.length; _c++) {
            var _d = propEntries_2[_c], key = _d[0], type = _d[1];
            var value = input[key];
            var outputValue = void 0;
            try {
                outputValue = type.parse(value, options);
            }
            catch (error) {
                issues = raiseIssuesOrCaptureForKey(error, issues, options, key);
            }
            if (isEqual(outputValue, value) || issues !== undefined) {
                continue;
            }
            if (output === input) {
                output = copyObjectEnumerableKeys(input);
            }
            output[key] = outputValue;
        }
        if (keysMode === 3 /* ObjectKeysMode.INDEXER */) {
            for (var key in input) {
                if (keys.includes(key)) {
                    continue;
                }
                var value = input[key];
                var outputValue = void 0;
                try {
                    outputValue = indexerType.parse(value, options);
                }
                catch (error) {
                    issues = raiseIssuesOrCaptureForKey(error, issues, options, key);
                }
                if (isEqual(outputValue, value) || issues !== undefined) {
                    continue;
                }
                if (output === input) {
                    output = copyObjectEnumerableKeys(input);
                }
                output[key] = outputValue;
            }
        }
        raiseIssuesIfDefined(issues);
        return output;
    };
    return ObjectType;
}(Type));

/**
 * The type definition that constrains input to one of the primitive values.
 *
 * @template T The type of the allowed values.
 */
var OneOfType = /** @class */ (function (_super) {
    tslib.__extends(OneOfType, _super);
    /**
     * Creates a new {@link OneOfType} instance.
     *
     * @param values The list of values allowed for the input.
     * @param options
     */
    function OneOfType(values, options) {
        var _this = _super.call(this, false, options) || this;
        _this.values = values;
        return _this;
    }
    OneOfType.prototype.parse = function (input, options) {
        var values = this.values;
        if (!values.includes(input)) {
            raiseIssue(input, 'oneOf', values, this.options, 'Must be equal to one of: ' + values.join(', '));
        }
        return input;
    };
    return OneOfType;
}(Type));

/**
 * The optional type definition.
 *
 * @template X The underlying type definition.
 */
var OptionalType = /** @class */ (function (_super) {
    tslib.__extends(OptionalType, _super);
    /**
     * Creates a new {@link OptionalType} instance.
     *
     * @param type The underlying type definition.
     * @param defaultValue The value that should be used if input is `undefined`.
     */
    function OptionalType(type, defaultValue) {
        var _this = _super.call(this, type.async) || this;
        _this.type = type;
        _this.defaultValue = defaultValue;
        return _this;
    }
    OptionalType.prototype.parse = function (input, options) {
        var _a = this, type = _a.type, defaultValue = _a.defaultValue;
        if (input === undefined) {
            return type.async ? Promise.resolve(defaultValue) : defaultValue;
        }
        return type.parse(input, options);
    };
    return OptionalType;
}(Type));

var RecordType = /** @class */ (function (_super) {
    tslib.__extends(RecordType, _super);
    function RecordType(keyType, valueType, options) {
        var _this = _super.call(this, keyType.async || valueType.async, options) || this;
        _this.keyType = keyType;
        _this.valueType = valueType;
        return _this;
    }
    RecordType.prototype.parse = function (input, options) {
        if (!isObjectLike(input)) {
            raiseIssue(input, 'type', 'object', this.options, 'Must be an object');
        }
        var _a = this, keyType = _a.keyType, valueType = _a.valueType;
        if (this.async) {
            var results = [];
            var handleResults = function (results) {
                var output = input;
                var i = 0;
                for (var key in input) {
                    var outputKey = results[i];
                    var outputValue = results[i + 1];
                    if (key === outputKey && isEqual(input[key], outputValue) && output === input) {
                        continue;
                    }
                    if (output === input) {
                        output = copyObjectEnumerableKeys(input, i);
                    }
                    output[outputKey] = outputValue;
                    i += 2;
                }
                return output;
            };
            for (var key in input) {
                results.push(parseAsync(keyType, key, options).catch(createCatchForKey(key)), parseAsync(valueType, input[key], options).catch(createCatchForKey(key)));
            }
            if (isFast(options)) {
                return promiseAll(results).then(handleResults);
            }
            return promiseAllSettled(results).then(createValuesExtractor()).then(handleResults);
        }
        var output = input;
        var issues;
        var i = 0;
        for (var key in input) {
            ++i;
            var value = input[key];
            var outputKey = key;
            var outputValue = void 0;
            try {
                outputKey = keyType.parse(key, options);
            }
            catch (error) {
                issues = raiseIssuesOrCaptureForKey(error, issues, options, key);
            }
            try {
                outputValue = valueType.parse(value, options);
            }
            catch (error) {
                issues = raiseIssuesOrCaptureForKey(error, issues, options, key);
            }
            if ((key === outputKey && isEqual(value, outputValue) && output === input) || issues !== undefined) {
                continue;
            }
            if (output === input) {
                output = copyObjectEnumerableKeys(input, i);
            }
            output[outputKey] = outputValue;
        }
        raiseIssuesIfDefined(issues);
        return output;
    };
    return RecordType;
}(Type));

/**
 * The string type definition.
 */
var StringType = /** @class */ (function (_super) {
    tslib.__extends(StringType, _super);
    function StringType(options) {
        return _super.call(this, false, options) || this;
    }
    /**
     * Constrains the string length.
     */
    StringType.prototype.length = function (length, options) {
        return this.min(length, options).max(length, options);
    };
    /**
     * Constrains the string length to be greater than or equal to the length.
     */
    StringType.prototype.min = function (length, options) {
        var type = cloneObject(this);
        type.minLength = length;
        type.minLengthOptions = options;
        return type;
    };
    /**
     * Constrains the string length to be less than or equal to the length.
     */
    StringType.prototype.max = function (length, options) {
        var type = cloneObject(this);
        type.maxLength = length;
        type.maxLengthOptions = options;
        return type;
    };
    /**
     * Constrains the string to match a regexp.
     */
    StringType.prototype.regex = function (re, options) {
        var type = cloneObject(this);
        type.re = re;
        type.reOptions = options;
        return type;
    };
    StringType.prototype.parse = function (input, options) {
        if (typeof input !== 'string') {
            raiseIssue(input, 'type', 'string', this.options, 'Must be a string');
        }
        var _a = this, minLength = _a.minLength, maxLength = _a.maxLength, re = _a.re;
        var issues;
        if (minLength != null && input.length < minLength) {
            issues = raiseIssuesOrPush(issues, options, input, 'stringMinLength', minLength, this.minLengthOptions, 'Must have the minimum length of ' + minLength);
        }
        if (maxLength != null && input.length > maxLength) {
            issues = raiseIssuesOrPush(issues, options, input, 'stringMaxLength', maxLength, this.maxLengthOptions, 'Must have the maximum length of ' + maxLength);
        }
        if (re != null && !re.test(input)) {
            issues = raiseIssuesOrPush(issues, options, input, 'stringRegex', re, this.reOptions, 'Must match the pattern ' + re);
        }
        raiseIssuesIfDefined(issues);
        return input;
    };
    return StringType;
}(Type));

/**
 * The tuple type definition.
 *
 * @template U The list of tuple elements.
 */
var TupleType = /** @class */ (function (_super) {
    tslib.__extends(TupleType, _super);
    /**
     * Creates a new {@link TupleType} instance.
     *
     * @param types The list of tuple elements.
     * @param options
     */
    function TupleType(types, options) {
        var _this = _super.call(this, isAsync(types), options) || this;
        _this.types = types;
        return _this;
    }
    TupleType.prototype.parse = function (input, options) {
        if (!isArray(input)) {
            raiseIssue(input, 'type', 'array', this.options, 'Must be an array');
        }
        var types = this.types;
        var typesLength = types.length;
        if (input.length !== typesLength) {
            raiseIssue(input, 'tupleLength', typesLength, this.options, 'Must have a length of ' + typesLength);
        }
        if (this.async) {
            var promises = [];
            var handleOutput = function (output) { return (isEqualArray(input, output) ? input : output); };
            for (var i = 0; i < typesLength; ++i) {
                promises.push(parseAsync(types[i], input[i], options).catch(createCatchForKey(i)));
            }
            if (isFast(options)) {
                return promiseAll(promises).then(handleOutput);
            }
            return promiseAllSettled(promises).then(createValuesExtractor()).then(handleOutput);
        }
        var output = input;
        var issues;
        for (var i = 0; i < typesLength; ++i) {
            var value = input[i];
            var outputValue = void 0;
            try {
                outputValue = types[i].parse(value, options);
            }
            catch (error) {
                issues = raiseIssuesOrCaptureForKey(error, issues, options, i);
            }
            if (isEqual(outputValue, value) || issues !== undefined) {
                continue;
            }
            if (output === input) {
                output = input.slice(0);
            }
            output[i] = outputValue;
        }
        raiseIssuesIfDefined(issues);
        return output;
    };
    return TupleType;
}(Type));

/**
 * The unconstrained type definition.
 *
 * @template The type of the unconstrained value.
 */
var UnconstrainedType = /** @class */ (function (_super) {
    tslib.__extends(UnconstrainedType, _super);
    function UnconstrainedType() {
        return _super.call(this, false) || this;
    }
    UnconstrainedType.prototype.parse = function (input, options) {
        return input;
    };
    return UnconstrainedType;
}(Type));

/**
 * The union type definition.
 *
 * @template U The list of united type definitions.
 */
var UnionType = /** @class */ (function (_super) {
    tslib.__extends(UnionType, _super);
    /**
     * Creates a new {@link UnionType} instance.
     *
     * @param types The list of united type definitions.
     */
    function UnionType(types) {
        var _this = _super.call(this, isAsync(types)) || this;
        _this.types = types;
        return _this;
    }
    UnionType.prototype.parse = function (input, options) {
        var _this = this;
        var types = this.types;
        if (this.async) {
            var promises = [];
            for (var _i = 0, types_1 = types; _i < types_1.length; _i++) {
                var type = types_1[_i];
                promises.push(parseAsync(type, input, options));
            }
            return promiseAllSettled(promises).then(function (results) {
                var issues;
                for (var _i = 0, results_1 = results; _i < results_1.length; _i++) {
                    var result = results_1[_i];
                    if (result.status === 'fulfilled') {
                        return result.value;
                    }
                    issues = extractIssues(result.reason);
                }
                raiseIssue(input, 'union', issues, _this.options, 'Must conform a union');
            });
        }
        var issues;
        for (var _a = 0, types_2 = types; _a < types_2.length; _a++) {
            var type = types_2[_a];
            try {
                return type.parse(input);
            }
            catch (error) {
                issues = extractIssues(error);
            }
        }
        raiseIssue(input, 'union', issues, this.options, 'Must conform a union');
    };
    return UnionType;
}(Type));

/**
 * Creates the unconstrained type definition that accepts any value.
 */
function any() {
    return new UnconstrainedType();
}

/**
 * Creates the array type definition.
 *
 * @param type The type definition of array elements.
 * @param options
 */
function array(type, options) {
    return new ArrayType(type, options);
}

/**
 * Creates the bigint type definition.
 */
function bigint(options) {
    return new BigIntType(options);
}

/**
 * Creates the boolean type definition.
 */
function boolean(options) {
    return new BooleanType(options);
}

/**
 * Creates the class instance type definition.
 *
 * @template C The type of the instance constructor.
 */
function instanceOf(ctor, options) {
    return new InstanceOfType(ctor, options);
}

/**
 * Creates the integer type definition.
 */
function integer(options) {
    return new IntegerType(options);
}

/**
 * Returns the type definition that is lazily resolved at parse time.
 *
 * @param provider Returns the type definition that must be applied to the input.
 *
 * @template X The type definition returned by the provider.
 */
function lazy(provider) {
    return new LazyType(false, provider);
}

/**
 * Creates the literal value type definition.
 *
 * @param value The literal value to which the value must be equal.
 * @param options
 *
 * @template T The type of the literal value.
 */
function literal(value, options) {
    return new LiteralType(value, options);
}

/**
 * Creates the type definition that always raises an issue.
 */
function never(options) {
    return new NeverType(options);
}

/**
 * Creates the nullable type definition.
 *
 * @param type The underlying type definition.
 */
function nullable(type) {
    return new NullableType(type);
}

/**
 * Marks the type as optional.
 *
 * @param type The type that must be optional.
 * @param defaultValue The value that should be used if input is `undefined`.
 */
function optional(type, defaultValue) {
    return new OptionalType(type, defaultValue);
}

/**
 * Creates the type definition that allows both `undefined` and `null` values.
 *
 * @param type The underlying type definition.
 * @param defaultValue The value that should be used if input is `undefined`.
 */
function nullish(type, defaultValue) {
    return optional(nullable(type), defaultValue);
}

/**
 * Creates the number type definition.
 */
function number(options) {
    return new NumberType(options);
}

/**
 * Creates the array type definition.
 *
 * @param props The mapping from an object key to a corresponding type definition.
 * @param options
 */
function object(props, options) {
    return new ObjectType(props, null, options);
}

function oneOf(values, options) {
    if (Array.isArray(values)) {
        return new OneOfType(values, options);
    }
    var enumValues = [];
    for (var _i = 0, _a = Object.values(values); _i < _a.length; _i++) {
        var value = _a[_i];
        if (typeof values[value] !== 'number') {
            enumValues.push(value);
        }
    }
    return new OneOfType(enumValues, options);
}

/**
 * Creates a union type definition that tries to parse the input with one of the provided types.
 *
 * @param types The list of types to try.
 */
function or(types) {
    return new UnionType(types);
}

function record(keyType, valueType, options) {
    return new RecordType(keyType, valueType, options);
}

function string(options) {
    return new StringType(options);
}

/**
 * Creates the tuple type definition.
 *
 * @param types The list of tuple elements.
 * @param options
 */
function tuple(types, options) {
    return new TupleType(types, options);
}

function unknown() {
    return new UnconstrainedType();
}

exports.ArrayType = ArrayType;
exports.BigIntType = BigIntType;
exports.BooleanType = BooleanType;
exports.InstanceOfType = InstanceOfType;
exports.IntegerType = IntegerType;
exports.LazyType = LazyType;
exports.LiteralType = LiteralType;
exports.NeverType = NeverType;
exports.NullableType = NullableType;
exports.NumberType = NumberType;
exports.ObjectType = ObjectType;
exports.OneOfType = OneOfType;
exports.OptionalType = OptionalType;
exports.RecordType = RecordType;
exports.StringType = StringType;
exports.TransformedType = TransformedType;
exports.TupleType = TupleType;
exports.Type = Type;
exports.UnconstrainedType = UnconstrainedType;
exports.UnionType = UnionType;
exports.ValidationError = ValidationError;
exports.any = any;
exports.array = array;
exports.bigint = bigint;
exports.boolean = boolean;
exports.instanceOf = instanceOf;
exports.integer = integer;
exports.lazy = lazy;
exports.literal = literal;
exports.never = never;
exports.nullable = nullable;
exports.nullish = nullish;
exports.number = number;
exports.object = object;
exports.oneOf = oneOf;
exports.optional = optional;
exports.or = or;
exports.record = record;
exports.string = string;
exports.tuple = tuple;
exports.unknown = unknown;
