import { AnyType, InferType, Type } from './Type';
import { Awaitable, ConstraintOptions, ParserOptions } from '../shared-types';
/**
 * The array type definition.
 *
 * @template X The type definition of array elements.
 */
export declare class ArrayType<X extends AnyType> extends Type<InferType<X>[]> {
    protected type: X;
    protected minLength?: number;
    protected maxLength?: number;
    protected minLengthOptions?: ConstraintOptions;
    protected maxLengthOptions?: ConstraintOptions;
    /**
     * Creates the new {@link ArrayType} instance.
     *
     * @param type The type definition of array elements.
     * @param options The type constraint options.
     */
    constructor(type: X, options?: ConstraintOptions);
    /**
     * Constrains the array length.
     *
     * @param length The length of the array to satisfy the constraint.
     * @param options The options applied to both {@link min} and {@link max} constraints.
     * @returns The new type definition.
     */
    length(length: number, options?: ConstraintOptions): this;
    /**
     * Constrains the minimum array length.
     *
     * @param length The minimum length of the array to satisfy the constraint.
     * @param options The minimum constraint options.
     * @returns The new type definition.
     */
    min(length: number, options?: ConstraintOptions): this;
    /**
     * Constrains the maximum array length.
     *
     * @param length The maximum length of the array to satisfy the constraint.
     * @param options The maximum constraint options.
     * @returns The new type definition.
     */
    max(length: number, options?: ConstraintOptions): this;
    parse(input: unknown, options?: ParserOptions): Awaitable<InferType<X>[]>;
}
