import { Type } from './Type';
import { ConstraintOptions, ParserOptions } from '../shared-types';
/**
 * The number type definition.
 */
export declare class NumberType extends Type<number> {
    protected min?: number;
    protected max?: number;
    protected minIncluded?: boolean;
    protected maxIncluded?: boolean;
    protected divisor?: number;
    protected minOptions?: ConstraintOptions;
    protected maxOptions?: ConstraintOptions;
    protected divisorOptions?: ConstraintOptions;
    constructor(options?: ConstraintOptions);
    /**
     * Constrains the number to be greater than zero.
     */
    positive(options?: ConstraintOptions): this;
    /**
     * Constrains the number to be less than zero.
     */
    negative(options?: ConstraintOptions): this;
    /**
     * Constrains the number to be greater than the value.
     */
    gt(value: number, options?: ConstraintOptions): this;
    /**
     * Constrains the number to be less than the value.
     */
    lt(value: number, options?: ConstraintOptions): this;
    /**
     * Constrains the number to be greater than or equal to the value.
     */
    gte(value: number, options?: ConstraintOptions): this;
    /**
     * Constrains the number to be less than or equal to the value.
     */
    lte(value: number, options?: ConstraintOptions): this;
    /**
     * Constrains the number to be a multiple of the divisor.
     *
     * @param divisor The number by which the input should be divisible without a remainder.
     * @param options
     */
    multipleOf(divisor: number, options?: ConstraintOptions): this;
    parse(input: any, options?: ParserOptions): number;
}
