import { __extends } from "tslib";
import { Type } from './Type';
import { raiseIssue } from '../utils';
/**
 * The type definition that constrains input to one of the primitive values.
 *
 * @template T The type of the allowed values.
 */
var OneOfType = /** @class */ (function (_super) {
    __extends(OneOfType, _super);
    /**
     * Creates a new {@link OneOfType} instance.
     *
     * @param values The list of values allowed for the input.
     * @param options
     */
    function OneOfType(values, options) {
        var _this = _super.call(this, false, options) || this;
        _this.values = values;
        return _this;
    }
    OneOfType.prototype.parse = function (input, options) {
        var values = this.values;
        if (!values.includes(input)) {
            raiseIssue(input, 'oneOf', values, this.options, 'Must be equal to one of: ' + values.join(', '));
        }
        return input;
    };
    return OneOfType;
}(Type));
export { OneOfType };
