import { AnyType, InferType, Type } from './Type';
import { Awaitable, ParserOptions } from '../shared-types';
/**
 * The optional type definition.
 *
 * @template X The underlying type definition.
 */
export declare class OptionalType<X extends AnyType> extends Type<InferType<X> | undefined> {
    protected type: X;
    protected defaultValue?: InferType<X> | undefined;
    /**
     * Creates a new {@link OptionalType} instance.
     *
     * @param type The underlying type definition.
     * @param defaultValue The value that should be used if input is `undefined`.
     */
    constructor(type: X, defaultValue?: InferType<X> | undefined);
    parse(input: unknown, options?: ParserOptions): Awaitable<InferType<X> | undefined>;
}
