import { __extends } from "tslib";
import { Type } from './Type';
import { copyObjectEnumerableKeys, createCatchForKey, createValuesExtractor, isEqual, isFast, isObjectLike, parseAsync, promiseAll, promiseAllSettled, raiseIssue, raiseIssuesIfDefined, raiseIssuesOrCaptureForKey, } from '../utils';
var RecordType = /** @class */ (function (_super) {
    __extends(RecordType, _super);
    function RecordType(keyType, valueType, options) {
        var _this = _super.call(this, keyType.async || valueType.async, options) || this;
        _this.keyType = keyType;
        _this.valueType = valueType;
        return _this;
    }
    RecordType.prototype.parse = function (input, options) {
        if (!isObjectLike(input)) {
            raiseIssue(input, 'type', 'object', this.options, 'Must be an object');
        }
        var _a = this, keyType = _a.keyType, valueType = _a.valueType;
        if (this.async) {
            var results = [];
            var handleResults = function (results) {
                var output = input;
                var i = 0;
                for (var key in input) {
                    var outputKey = results[i];
                    var outputValue = results[i + 1];
                    if (key === outputKey && isEqual(input[key], outputValue) && output === input) {
                        continue;
                    }
                    if (output === input) {
                        output = copyObjectEnumerableKeys(input, i);
                    }
                    output[outputKey] = outputValue;
                    i += 2;
                }
                return output;
            };
            for (var key in input) {
                results.push(parseAsync(keyType, key, options).catch(createCatchForKey(key)), parseAsync(valueType, input[key], options).catch(createCatchForKey(key)));
            }
            if (isFast(options)) {
                return promiseAll(results).then(handleResults);
            }
            return promiseAllSettled(results).then(createValuesExtractor()).then(handleResults);
        }
        var output = input;
        var issues;
        var i = 0;
        for (var key in input) {
            ++i;
            var value = input[key];
            var outputKey = key;
            var outputValue = void 0;
            try {
                outputKey = keyType.parse(key, options);
            }
            catch (error) {
                issues = raiseIssuesOrCaptureForKey(error, issues, options, key);
            }
            try {
                outputValue = valueType.parse(value, options);
            }
            catch (error) {
                issues = raiseIssuesOrCaptureForKey(error, issues, options, key);
            }
            if ((key === outputKey && isEqual(value, outputValue) && output === input) || issues !== undefined) {
                continue;
            }
            if (output === input) {
                output = copyObjectEnumerableKeys(input, i);
            }
            output[outputKey] = outputValue;
        }
        raiseIssuesIfDefined(issues);
        return output;
    };
    return RecordType;
}(Type));
export { RecordType };
