import { Type } from './Type';
import { ConstraintOptions, ParserOptions } from '../shared-types';
/**
 * The string type definition.
 */
export declare class StringType extends Type<string> {
    protected minLength?: number;
    protected maxLength?: number;
    protected re?: RegExp;
    protected minLengthOptions?: ConstraintOptions;
    protected maxLengthOptions?: ConstraintOptions;
    protected reOptions?: ConstraintOptions;
    constructor(options?: ConstraintOptions);
    /**
     * Constrains the string length.
     */
    length(length: number, options?: ConstraintOptions): this;
    /**
     * Constrains the string length to be greater than or equal to the length.
     */
    min(length: number, options?: ConstraintOptions): this;
    /**
     * Constrains the string length to be less than or equal to the length.
     */
    max(length: number, options?: ConstraintOptions): this;
    /**
     * Constrains the string to match a regexp.
     */
    regex(re: RegExp, options?: ConstraintOptions): this;
    parse(input: unknown, options?: ParserOptions): string;
}
