import { __extends } from "tslib";
import { Type } from './Type';
import { cloneObject, raiseIssue, raiseIssuesIfDefined, raiseIssuesOrPush } from '../utils';
/**
 * The string type definition.
 */
var StringType = /** @class */ (function (_super) {
    __extends(StringType, _super);
    function StringType(options) {
        return _super.call(this, false, options) || this;
    }
    /**
     * Constrains the string length.
     */
    StringType.prototype.length = function (length, options) {
        return this.min(length, options).max(length, options);
    };
    /**
     * Constrains the string length to be greater than or equal to the length.
     */
    StringType.prototype.min = function (length, options) {
        var type = cloneObject(this);
        type.minLength = length;
        type.minLengthOptions = options;
        return type;
    };
    /**
     * Constrains the string length to be less than or equal to the length.
     */
    StringType.prototype.max = function (length, options) {
        var type = cloneObject(this);
        type.maxLength = length;
        type.maxLengthOptions = options;
        return type;
    };
    /**
     * Constrains the string to match a regexp.
     */
    StringType.prototype.regex = function (re, options) {
        var type = cloneObject(this);
        type.re = re;
        type.reOptions = options;
        return type;
    };
    StringType.prototype.parse = function (input, options) {
        if (typeof input !== 'string') {
            raiseIssue(input, 'type', 'string', this.options, 'Must be a string');
        }
        var _a = this, minLength = _a.minLength, maxLength = _a.maxLength, re = _a.re;
        var issues;
        if (minLength != null && input.length < minLength) {
            issues = raiseIssuesOrPush(issues, options, input, 'stringMinLength', minLength, this.minLengthOptions, 'Must have the minimum length of ' + minLength);
        }
        if (maxLength != null && input.length > maxLength) {
            issues = raiseIssuesOrPush(issues, options, input, 'stringMaxLength', maxLength, this.maxLengthOptions, 'Must have the maximum length of ' + maxLength);
        }
        if (re != null && !re.test(input)) {
            issues = raiseIssuesOrPush(issues, options, input, 'stringRegex', re, this.reOptions, 'Must match the pattern ' + re);
        }
        raiseIssuesIfDefined(issues);
        return input;
    };
    return StringType;
}(Type));
export { StringType };
