import { AnyType, InferType, Type } from './Type';
import { Awaitable, ConstraintOptions, ParserOptions, Several } from '../shared-types';
export declare type InferTupleType<U extends Several<AnyType>> = {
    [K in keyof U]: InferType<U[K]>;
};
/**
 * The tuple type definition.
 *
 * @template U The list of tuple elements.
 */
export declare class TupleType<U extends Several<AnyType>> extends Type<InferTupleType<U>> {
    protected types: U;
    /**
     * Creates a new {@link TupleType} instance.
     *
     * @param types The list of tuple elements.
     * @param options
     */
    constructor(types: U, options?: ConstraintOptions);
    parse(input: unknown, options?: ParserOptions): Awaitable<InferTupleType<U>>;
}
