import { __extends } from "tslib";
import { Type } from './Type';
import { createCatchForKey, createValuesExtractor, isArray, isAsync, isEqual, isEqualArray, isFast, parseAsync, promiseAll, promiseAllSettled, raiseIssue, raiseIssuesIfDefined, raiseIssuesOrCaptureForKey, } from '../utils';
/**
 * The tuple type definition.
 *
 * @template U The list of tuple elements.
 */
var TupleType = /** @class */ (function (_super) {
    __extends(TupleType, _super);
    /**
     * Creates a new {@link TupleType} instance.
     *
     * @param types The list of tuple elements.
     * @param options
     */
    function TupleType(types, options) {
        var _this = _super.call(this, isAsync(types), options) || this;
        _this.types = types;
        return _this;
    }
    TupleType.prototype.parse = function (input, options) {
        if (!isArray(input)) {
            raiseIssue(input, 'type', 'array', this.options, 'Must be an array');
        }
        var types = this.types;
        var typesLength = types.length;
        if (input.length !== typesLength) {
            raiseIssue(input, 'tupleLength', typesLength, this.options, 'Must have a length of ' + typesLength);
        }
        if (this.async) {
            var promises = [];
            var handleOutput = function (output) { return (isEqualArray(input, output) ? input : output); };
            for (var i = 0; i < typesLength; ++i) {
                promises.push(parseAsync(types[i], input[i], options).catch(createCatchForKey(i)));
            }
            if (isFast(options)) {
                return promiseAll(promises).then(handleOutput);
            }
            return promiseAllSettled(promises).then(createValuesExtractor()).then(handleOutput);
        }
        var output = input;
        var issues;
        for (var i = 0; i < typesLength; ++i) {
            var value = input[i];
            var outputValue = void 0;
            try {
                outputValue = types[i].parse(value, options);
            }
            catch (error) {
                issues = raiseIssuesOrCaptureForKey(error, issues, options, i);
            }
            if (isEqual(outputValue, value) || issues !== undefined) {
                continue;
            }
            if (output === input) {
                output = input.slice(0);
            }
            output[i] = outputValue;
        }
        raiseIssuesIfDefined(issues);
        return output;
    };
    return TupleType;
}(Type));
export { TupleType };
