import { __extends } from "tslib";
import { Type } from './Type';
import { extractIssues, isAsync, parseAsync, promiseAllSettled, raiseIssue } from '../utils';
/**
 * The union type definition.
 *
 * @template U The list of united type definitions.
 */
var UnionType = /** @class */ (function (_super) {
    __extends(UnionType, _super);
    /**
     * Creates a new {@link UnionType} instance.
     *
     * @param types The list of united type definitions.
     */
    function UnionType(types) {
        var _this = _super.call(this, isAsync(types)) || this;
        _this.types = types;
        return _this;
    }
    UnionType.prototype.parse = function (input, options) {
        var _this = this;
        var types = this.types;
        if (this.async) {
            var promises = [];
            for (var _i = 0, types_1 = types; _i < types_1.length; _i++) {
                var type = types_1[_i];
                promises.push(parseAsync(type, input, options));
            }
            return promiseAllSettled(promises).then(function (results) {
                var issues;
                for (var _i = 0, results_1 = results; _i < results_1.length; _i++) {
                    var result = results_1[_i];
                    if (result.status === 'fulfilled') {
                        return result.value;
                    }
                    issues = extractIssues(result.reason);
                }
                raiseIssue(input, 'union', issues, _this.options, 'Must conform a union');
            });
        }
        var issues;
        for (var _a = 0, types_2 = types; _a < types_2.length; _a++) {
            var type = types_2[_a];
            try {
                return type.parse(input);
            }
            catch (error) {
                issues = extractIssues(error);
            }
        }
        raiseIssue(input, 'union', issues, this.options, 'Must conform a union');
    };
    return UnionType;
}(Type));
export { UnionType };
