import { ValidationError } from './ValidationError';
export function raiseIssue(input, code, param, options, message) {
    var _a;
    raiseIssuesIfDefined([
        {
            code: code,
            path: [],
            input: input,
            param: param,
            message: (_a = options === null || options === void 0 ? void 0 : options.message) !== null && _a !== void 0 ? _a : message,
            meta: options === null || options === void 0 ? void 0 : options.meta,
        },
    ]);
}
export function raiseIssuesOrPush(issues, parserOptions, input, code, param, options, message) {
    var _a;
    if (issues === void 0) { issues = []; }
    issues.push({
        code: code,
        path: [],
        input: input,
        param: param,
        message: (_a = options === null || options === void 0 ? void 0 : options.message) !== null && _a !== void 0 ? _a : message,
        meta: options === null || options === void 0 ? void 0 : options.meta,
    });
    if (parserOptions === null || parserOptions === void 0 ? void 0 : parserOptions.fast) {
        raiseIssuesIfDefined(issues);
    }
    return issues;
}
export function raiseIssuesIfDefined(issues) {
    if (issues !== undefined) {
        throw new ValidationError(issues);
    }
}
export function raiseIssuesOrCaptureForKey(error, issues, parserOptions, key) {
    if (error instanceof ValidationError) {
        for (var _i = 0, _a = error.issues; _i < _a.length; _i++) {
            var issue = _a[_i];
            issue.path.unshift(key);
        }
    }
    if (isFast(parserOptions) || !(error instanceof ValidationError)) {
        throw error;
    }
    if (issues !== undefined) {
        issues.push.apply(issues, error.issues);
        return issues;
    }
    return error.issues;
}
export function cloneObject(value) {
    return Object.assign(Object.create(Object.getPrototypeOf(value)), value);
}
export function isObjectLike(value) {
    return value !== null && typeof value === 'object';
}
export function createCatchForKey(key) {
    return function (error) {
        if (error instanceof ValidationError) {
            for (var _i = 0, _a = error.issues; _i < _a.length; _i++) {
                var issue = _a[_i];
                issue.path.unshift(key);
            }
        }
        throw error;
    };
}
export function createValuesExtractor(issues) {
    return function (results) {
        for (var i = 0; i < results.length; ++i) {
            var result = results[i];
            if (result.status === 'rejected') {
                var error = result.reason;
                if (!(error instanceof ValidationError)) {
                    throw error;
                }
                if (issues === undefined) {
                    issues = error.issues;
                }
                else {
                    issues.push.apply(issues, error.issues);
                }
            }
        }
        if (issues !== undefined) {
            throw new ValidationError(issues);
        }
        var values = [];
        for (var i = 0; i < results.length; ++i) {
            values.push(results[i].value);
        }
        return values;
    };
}
export function isFast(options) {
    return options !== undefined && options.fast === true;
}
export function isAsync(types) {
    var async = false;
    for (var i = 0; i < types.length && !async; ++i) {
        async = types[i].async;
    }
    return async;
}
export var isEqual = Object.is;
export var isArray = Array.isArray;
export var promiseAll = Promise.all.bind(Promise);
export var promiseAllSettled = Promise.allSettled.bind(Promise);
export function parseAsync(type, input, options) {
    return type.async ? type.parse(input, options) : Promise.resolve().then(function () { return type.parse(input, options); });
}
export function isEqualArray(a, b) {
    for (var i = 0; i < a.length; ++i) {
        if (!isEqual(a[i], b[i])) {
            return false;
        }
    }
    return true;
}
export function returnNull() {
    return null;
}
export function extractIssues(error) {
    if (error instanceof ValidationError) {
        return error.issues;
    }
    throw error;
}
export function copyObjectKnownKeys(input, keys) {
    var output = {};
    for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
        var key = keys_1[_i];
        output[key] = input[key];
    }
    return output;
}
export function copyObjectEnumerableKeys(input, keyCount) {
    var output = {};
    if (keyCount === undefined) {
        for (var key in input) {
            output[key] = input[key];
        }
        return output;
    }
    var i = 0;
    for (var key in input) {
        if (i >= keyCount) {
            break;
        }
        output[key] = input[key];
        i++;
    }
    return output;
}
export function returnTrue() {
    return true;
}
export function returnFalse() {
    return false;
}
